@@ -1,15 +1,25 @@
+/* $Id: local.c,v 1.13 2001/12/03 18:29:37 ukai Exp $ */
 #include "fm.h"
 #include <string.h>
 #include <stdio.h>
 #include <stdlib.h>
-#ifdef READLINK
+#include <sys/types.h>
+#include <signal.h>
+#include <errno.h>
+#ifdef HAVE_READLINK
 #include <unistd.h>
-#endif				/* READLINK */
+#endif				/* HAVE_READLINK */
+#ifdef __EMX__
+#include <limits.h>		/* _MAX_PATH ? */
+#endif				/* __EMX__ */
 #include "local.h"
 
 #define CGIFN_NORMAL     0
 #define CGIFN_DROOT      1
 #define CGIFN_CGIBIN     2
+#define CGIFN_MODE(x) ((x)&3)
+
+#define CGIFN_CONTAIN_SLASH 4
 
 /* setup cookie for local CGI */
 void
@@ -17,8 +27,9 @@ setLocalCookie()
 {
     Str buf;
     char hostname[256];
-    gethostname(hostname,256);
-    buf = Sprintf("%d.%ld@%s",getpid(),lrand48(),hostname);
+    gethostname(hostname, 256);
+
+    buf = Sprintf("%d.%ld@%s", getpid(), lrand48(), hostname);
     Local_cookie = buf->ptr;
 }
 
@@ -32,24 +43,28 @@ dirBuffer(char *dname)
     char **flist;
     char *p, *qdir;
     Str fbuf = Strnew();
-#ifdef READLINK
+#ifdef HAVE_READLINK
     struct stat lst;
     char lbuf[1024];
-#endif				/* READLINK */
-    int i, l, nrow, n = 0, maxlen = 0;
+#endif				/* HAVE_READLINK */
+    int i, l, nrow = 0, n = 0, maxlen = 0;
     int nfile, nfile_max = 100;
     Str dirname;
+    Buffer *buf;
 
     d = opendir(dname);
     if (d == NULL)
 	return NULL;
     dirname = Strnew_charp(dname);
-    qdir = htmlquote_str(dirname->ptr);
-    tmp = Sprintf("<title>Directory list of %s</title><h1>Directory list of %s</h1>\n", qdir, qdir);
+    qdir = html_quote(Str_conv_from_system(dirname)->ptr);
+    tmp =
+	Sprintf
+	("<title>Directory list of %s</title><h1>Directory list of %s</h1>\n",
+	 qdir, qdir);
     flist = New_N(char *, nfile_max);
     nfile = 0;
     while ((dir = readdir(d)) != NULL) {
-	flist[nfile++] = allocStr(dir->d_name, 0);
+	flist[nfile++] = allocStr(dir->d_name, -1);
 	if (nfile == nfile_max) {
 	    nfile_max *= 2;
 	    flist = New_Reuse(char *, flist, nfile_max);
@@ -70,7 +85,7 @@ dirBuffer(char *dname)
 	n = 1;
 	Strcat_charp(tmp, "<TABLE CELLPADDING=0><TR VALIGN=TOP>\n");
     }
-    qsort((void *) flist, nfile, sizeof(char *), strCmp);
+    qsort((void *)flist, nfile, sizeof(char *), strCmp);
     for (i = 0; i < nfile; i++) {
 	p = flist[i];
 	if (strcmp(p, ".") == 0)
@@ -79,7 +94,7 @@ dirBuffer(char *dname)
 	if (Strlastchar(fbuf) != '/')
 	    Strcat_char(fbuf, '/');
 	Strcat_charp(fbuf, p);
-#ifdef READLINK
+#ifdef HAVE_READLINK		/* readlink == lstat() ? (ukai) */
 	if (lstat(fbuf->ptr, &lst) < 0)
 	    continue;
 #endif				/* READLINK */
@@ -92,17 +107,17 @@ dirBuffer(char *dname)
 	else {
 	    if (S_ISDIR(st.st_mode))
 		Strcat_charp(tmp, "[DIR]&nbsp; ");
-#ifdef READLINK
+#ifdef HAVE_READLINK
 	    else if (S_ISLNK(lst.st_mode))
 		Strcat_charp(tmp, "[LINK] ");
-#endif				/* READLINE */
+#endif				/* HAVE_READLINE */
 	    else
 		Strcat_charp(tmp, "[FILE] ");
 	}
-	Strcat_m_charp(tmp, "<A HREF=\"", htmlquote_str(fbuf->ptr), NULL);
+	Strcat_m_charp(tmp, "<A HREF=\"", file_to_url(fbuf->ptr), NULL);
 	if (S_ISDIR(st.st_mode))
 	    Strcat_char(tmp, '/');
-	Strcat_m_charp(tmp, "\">", htmlquote_str(p), NULL);
+	Strcat_m_charp(tmp, "\">", html_quote(conv_from_system(p)), NULL);
 	if (S_ISDIR(st.st_mode))
 	    Strcat_char(tmp, '/');
 	Strcat_charp(tmp, "</a>");
@@ -116,16 +131,17 @@ dirBuffer(char *dname)
 	    }
 	}
 	else {
-#ifdef READLINK
+#ifdef HAVE_READLINK
 	    if (S_ISLNK(lst.st_mode)) {
 		if ((l = readlink(fbuf->ptr, lbuf, sizeof(lbuf))) > 0) {
 		    lbuf[l] = '\0';
-		    Strcat_m_charp(tmp, " -> ", htmlquote_str(lbuf), NULL);
+		    Strcat_m_charp(tmp, " -> ",
+				   html_quote(conv_from_system(lbuf)), NULL);
 		    if (S_ISDIR(st.st_mode))
 			Strcat_char(tmp, '/');
 		}
 	    }
-#endif				/* READLINK */
+#endif				/* HAVE_READLINK */
 	    Strcat_charp(tmp, "<br>\n");
 	}
     }
@@ -133,52 +149,56 @@ dirBuffer(char *dname)
 	Strcat_charp(tmp, "</TR></TABLE>\n");
     }
 
-    return loadHTMLString(tmp);
+    buf = loadHTMLString(tmp);
+#ifdef JP_CHARSET
+    if (buf)
+	buf->document_code = SystemCode;
+#endif
+    return buf;
 }
 
-#ifdef __EMX__
-char *
-get_os2_dft(const char *name, char *dft)
-{
-    char *value = getenv(name);
-    return value ? value : dft;
-}
-
-#define lib_dir get_os2_dft("W3M_LIB_DIR",LIB_DIR)
-#else				/* not __EMX__ */
-#define lib_dir LIB_DIR
-#endif				/* not __EMX__ */
-
 static int
 check_local_cgi(char *file, int status)
 {
     struct stat st;
 
+    if (status & CGIFN_CONTAIN_SLASH) {
+	/* local CGI file must be just under /cgi-bin/
+	 * or /$LIB/
+	 */
+	return -1;
+    }
 #ifdef __EMX__
-    if (status != CGIFN_CGIBIN) {
+    if (CGIFN_MODE(status) != CGIFN_CGIBIN) {
 	char tmp[_MAX_PATH];
+	int len;
 
-	_abspath(tmp, lib_dir, _MAX_PATH);	/* Translate '\\'  to  '/' 
-						 * 
-						 */
-	if (strnicmp(file, tmp, strlen(tmp)))	/* and ignore case  */
+	_abspath(tmp, w3m_lib_dir(), _MAX_PATH);	/* Translate '\\' to '/' */
+	len = strlen(tmp);
+	while (len > 1 && tmp[len - 1] == '/')
+	    len--;
+	if (strnicmp(file, tmp, len) ||	/* and ignore case  */
+	    (file[len] != '/'))
 	    return -1;
     }
 #else				/* not __EMX__ */
-    if (status != CGIFN_CGIBIN &&
-	strncmp(file, lib_dir, strlen(lib_dir)) != 0) {
-	/* 
-	 * a local-CGI script should be located on either
-	 * /cgi-bin/ directory or LIB_DIR (typically /usr/local/lib/w3m).
-	 */
-	return -1;
+    if (CGIFN_MODE(status) != CGIFN_CGIBIN) {
+	char *tmp = Strnew_charp(w3m_lib_dir())->ptr;
+	int len = strlen(tmp);
+
+	while (len > 1 && tmp[len - 1] == '/')
+	    len--;
+	if (strncmp(file, tmp, len) || (file[len] != '/'))
+	    /* 
+	     * a local-CGI script should be located on either
+	     * /cgi-bin/ directory or LIB_DIR (typically /usr/local/lib/w3m).
+	     */
+	    return -1;
     }
 #endif				/* not __EMX__ */
     if (stat(file, &st) < 0)
 	return -1;
-    if ((st.st_uid == geteuid() && (st.st_mode & S_IXUSR)) ||
-	(st.st_gid == getegid() && (st.st_mode & S_IXGRP)) ||
-	(st.st_mode & S_IXOTH)) {	/* executable */
+    if ((st.st_uid == geteuid() && (st.st_mode & S_IXUSR)) || (st.st_gid == getegid() && (st.st_mode & S_IXGRP)) || (st.st_mode & S_IXOTH)) {	/* executable */
 	return 0;
     }
     return -1;
@@ -188,7 +208,8 @@ void
 set_environ(char *var, char *value)
 {
 #ifdef HAVE_SETENV
-    setenv(var, value, 1);
+    if (var != NULL && value != NULL)
+	setenv(var, value, 1);
 #else				/* not HAVE_SETENV */
 #ifdef HAVE_PUTENV
     Str tmp = Strnew_m_charp(var, "=", value, NULL);
@@ -219,10 +240,10 @@ set_environ(char *var, char *value)
 	    }
 	}
     }
-    newenv = (char **) GC_malloc((i + 2) * sizeof(char *));
+    newenv = (char **)GC_malloc((i + 2) * sizeof(char *));
     if (newenv == NULL)
 	return;
-    for (e = environ, ne = newenv; *e != NULL; *(ne++) = *(e++));
+    for (e = environ, ne = newenv; *e != NULL; *(ne++) = *(e++)) ;
     *(ne++) = p;
     *ne = NULL;
     environ = newenv;
@@ -233,7 +254,7 @@ set_environ(char *var, char *value)
 static void
 set_cgi_environ(char *name, char *fn, char *req_uri)
 {
-    set_environ("SERVER_SOFTWARE", version);
+    set_environ("SERVER_SOFTWARE", w3m_version);
     set_environ("SERVER_PROTOCOL", "HTTP/1.0");
     set_environ("SERVER_NAME", "localhost");
     set_environ("SERVER_PORT", "80");	/* dummy */
@@ -244,7 +265,7 @@ set_cgi_environ(char *name, char *fn, char *req_uri)
     set_environ("SCRIPT_NAME", name);
     set_environ("SCRIPT_FILENAME", fn);
     set_environ("REQUEST_URI", req_uri);
-    set_environ("LOCAL_COOKIE",Local_cookie);
+    set_environ("LOCAL_COOKIE", Local_cookie);
 }
 
 static Str
@@ -274,6 +295,8 @@ cgi_filename(char *fn, int *status)
     struct stat st;
     if (cgi_bin != NULL && strncmp(fn, "/cgi-bin/", 9) == 0) {
 	*status = CGIFN_CGIBIN;
+	if (strchr(fn + 9, '/'))
+	    *status |= CGIFN_CONTAIN_SLASH;
 	tmp = checkPath(fn + 9, cgi_bin);
 	if (tmp == NULL)
 	    return fn;
@@ -281,8 +304,10 @@ cgi_filename(char *fn, int *status)
     }
     if (strncmp(fn, "/$LIB/", 6) == 0) {
 	*status = CGIFN_NORMAL;
-	tmp = Strnew_charp(lib_dir);
+	tmp = Strnew_charp(w3m_lib_dir());
 	fn += 5;
+	if (strchr(fn + 1, '/'))
+	    *status |= CGIFN_CONTAIN_SLASH;
 	if (Strlastchar(tmp) == '/')
 	    fn++;
 	Strcat_charp(tmp, fn);
@@ -290,6 +315,8 @@ cgi_filename(char *fn, int *status)
     }
     if (*fn == '/' && document_root != NULL && stat(fn, &st) < 0) {
 	*status = CGIFN_DROOT;
+	if (strchr(fn + 1, '/'))
+	    *status |= CGIFN_CONTAIN_SLASH;
 	tmp = Strnew_charp(document_root);
 	if (Strlastchar(tmp) != '/')
 	    Strcat_char(tmp, '/');
@@ -300,89 +327,155 @@ cgi_filename(char *fn, int *status)
     return fn;
 }
 
-FILE *
-localcgi_post(char *file, FormList * request, char *referer)
+static pid_t
+localcgi_popen_r(FILE ** p_fp)
 {
-    FILE *f;
-    Str tmp1, tmp2;
-    int status;
+    int fd[2];
+    FILE *fp;
+    pid_t pid;
+    Str emsg;
 
-    tmp1 = Strnew_charp(file);
-    file = cgi_filename(file, &status);
+    if (pipe(fd) < 0) {
+	emsg = Sprintf("localcgi_popen_r: pipe: %s", strerror(errno));
+	disp_err_message(emsg->ptr, FALSE);
+	return (pid_t) - 1;
+    }
+
+    flush_tty();
+    if ((pid = fork()) < 0) {
+	emsg = Sprintf("localcgi_popen_r: fork: %s", strerror(errno));
+	disp_err_message(emsg->ptr, FALSE);
+	close(fd[0]);
+	close(fd[1]);
+	return (pid_t) - 1;
+    }
+    else if (!pid) {
+	close_tty();
+	dup2(fd[1], 1);
+
+	if (fd[1] > 1)
+	    close(fd[1]);
+
+	close(fd[0]);
+    }
+    else {
+	close(fd[1]);
+
+	if (!(fp = fdopen(fd[0], "r"))) {
+	    emsg =
+		Sprintf("localcgi_popen_r: fdopen(%d, \"r\"): %s", fd[0],
+			strerror(errno));
+	    disp_err_message(emsg->ptr, FALSE);
+	    kill(pid, SIGTERM);
+	    close(fd[0]);
+	    return (pid_t) - 1;
+	}
+
+	*p_fp = fp;
+    }
+
+    return pid;
+}
+
+FILE *
+localcgi_post(char *uri, char *qstr, FormList *request, char *referer)
+{
+    FILE *f, *f1;
+    Str tmp1;
+    int status;
+    pid_t pid;
+    char *file;
+
+    file = cgi_filename(uri, &status);
     if (check_local_cgi(file, status) < 0)
 	return NULL;
-    set_cgi_environ(tmp1->ptr, file, tmp1->ptr);
+    tmp1 = tmpfname(TMPF_DFL, NULL);
+    f1 = fopen(tmp1->ptr, "w");
+    if (f1 == NULL)
+	return NULL;
+    pushText(fileToDelete, tmp1->ptr);
+    if ((pid = localcgi_popen_r(&f))) {
+	fclose(f1);
+	return pid > 0 ? f : NULL;
+    }
+    if (qstr == NULL) {
+	set_cgi_environ(uri, file, uri);
+    }
+    else {
+	set_cgi_environ(uri, file, Strnew_m_charp(uri, "?", qstr, NULL)->ptr);
+    }
     set_environ("REQUEST_METHOD", "POST");
+    if (qstr)
+	set_environ("QUERY_STRING", qstr);
     set_environ("CONTENT_LENGTH", Sprintf("%d", request->length)->ptr);
     if (referer && referer != NO_REFERER)
-        set_environ("HTTP_REFERER",referer);
+	set_environ("HTTP_REFERER", referer);
     if (request->enctype == FORM_ENCTYPE_MULTIPART) {
 	set_environ("CONTENT_TYPE",
-		    Sprintf("multipart/form-data; boundary=%s", request->boundary)->ptr);
+		    Sprintf("multipart/form-data; boundary=%s",
+			    request->boundary)->ptr);
     }
     else {
 	set_environ("CONTENT_TYPE", "application/x-www-form-urlencoded");
     }
-    tmp1 = tmpfname(TMPF_DFL, NULL);
-    f = fopen(tmp1->ptr, "w");
-    if (f == NULL)
-	return NULL;
-    pushText(fileToDelete, tmp1->ptr);
     if (request->enctype == FORM_ENCTYPE_MULTIPART) {
 	FILE *fd;
 	int c;
 	fd = fopen(request->body, "r");
 	if (fd != NULL) {
 	    while ((c = fgetc(fd)) != EOF)
-		fputc(c, f);
+		fputc(c, f1);
 	    fclose(fd);
 	}
     }
     else {
-	fputs(request->body, f);
+	fputs(request->body, f1);
     }
-    fclose(f);
-    tmp2 = Sprintf("%s < %s", file, tmp1->ptr);
-#ifdef __EMX__
-    f = popen(tmp2->ptr, "r");
-#else
-    tmp1 = Strnew_charp(CurrentDir);
+    fclose(f1);
+    freopen(tmp1->ptr, "r", stdin);
+#ifdef HAVE_CHDIR		/* ifndef __EMX__ ? */
     chdir(mydirname(file));
-    f = popen(tmp2->ptr, "r");
-    chdir(tmp1->ptr);
 #endif
-    return f;
+    execl(file, mybasename(file), NULL);
+    fprintf(stderr, "execl(\"%s\", \"%s\", NULL): %s\n",
+	    file, mybasename(file), strerror(errno));
+    exit(1);
+    return NULL;
 }
 
 FILE *
-localcgi_get(char *file, char *request, char *referer)
+localcgi_get(char *uri, char *request, char *referer)
 {
     FILE *f;
-    Str tmp1, tmp2;
     int status;
-    
-    tmp1 = Strnew_charp(file);
-    file = cgi_filename(file, &status);
+    pid_t pid;
+    char *file;
+
+    file = cgi_filename(uri, &status);
     if (check_local_cgi(file, status) < 0)
 	return NULL;
-    if (!strcmp(request, "")) {
-        set_cgi_environ(tmp1->ptr, file, tmp1->ptr);
-    } else {
-        set_cgi_environ(tmp1->ptr, file,
-			Strnew_m_charp(tmp1->ptr, "?", request, NULL)->ptr);
+    if ((pid = localcgi_popen_r(&f)) < 0)
+	return NULL;
+    else if (pid)
+	return f;
+    if (request == NULL) {
+	set_cgi_environ(Strnew_charp(uri)->ptr, file, Strnew_charp(uri)->ptr);
+    }
+    else {
+	set_cgi_environ(Strnew_charp(uri)->ptr, file,
+			Strnew_m_charp(uri, "?", request, NULL)->ptr);
     }
     if (referer && referer != NO_REFERER)
-        set_environ("HTTP_REFERER",referer);
+	set_environ("HTTP_REFERER", referer);
     set_environ("REQUEST_METHOD", "GET");
-    set_environ("QUERY_STRING", request);
-    tmp2 = Sprintf("%s", file);
-#ifdef __EMX__
-    f = popen(tmp2->ptr, "r");
-#else
-    tmp1 = Strnew_charp(CurrentDir);
+    set_environ("QUERY_STRING", request ? request : "");
+    freopen(DEV_NULL_PATH, "r", stdin);
+#ifdef HAVE_CHDIR		/* ifndef __EMX__? */
     chdir(mydirname(file));
-    f = popen(tmp2->ptr, "r");
-    chdir(tmp1->ptr);
 #endif
-    return f;
+    execl(file, mybasename(file), NULL);
+    fprintf(stderr, "execl(\"%s\", \"%s\", NULL): %s\n",
+	    file, mybasename(file), strerror(errno));
+    exit(1);
+    return NULL;
 }
