@@ -1,29 +1,33 @@
-/* $Id: mktable.c,v 1.5 2001/12/23 14:31:38 ukai Exp $ */
+/* $Id: mktable.c,v 1.8 2002/01/29 19:16:00 ukai Exp $ */
 #include "config.h"
 #include <stdio.h>
 #include <stdlib.h>
 #include <ctype.h>
-#ifdef HAVE_STDINT_H
-#include "stdint.h"
-#else
-typedef unsigned int uintptr_t;
-#endif
 #include "hash.h"
 #include "Str.h"
 #include <gc.h>
 
 #include "gcmain.c"
 
+/* *INDENT-OFF* */
 defhash(HashItem_ss *, int, hss_i)
+/* *INDENT-ON* */
+
 #define keycomp(x,y) ((x)==(y))
-     static uintptr_t hashfunc(HashItem_ss * x)
+
+/* XXX: we assume sizeof(unsigned long) >= sizeof(void *) */
+static unsigned long
+hashfunc(HashItem_ss * x)
 {
-    return (uintptr_t) x;
+    return (unsigned long)x;
 }
 
+/* *INDENT-OFF* */
 defhashfunc(HashItem_ss *, int, hss_i)
-     int
-      MAIN(int argc, char *argv[], char **envp)
+/* *INDENT-ON* */
+
+int
+MAIN(int argc, char *argv[], char **envp)
 {
     FILE *f;
     Hash_ss *hash;
