@@ -1,4 +1,4 @@
-/* $Id: url.c,v 1.28 2001/12/27 18:22:59 ukai Exp $ */
+/* $Id: url.c,v 1.39 2002/01/29 19:08:50 ukai Exp $ */
 #include "fm.h"
 #include <sys/types.h>
 #include <sys/socket.h>
@@ -69,7 +69,9 @@ struct cmdtable schemetable[] = {
     /*  {"exec", SCM_EXEC}, */
     {"nntp", SCM_NNTP},
     {"news", SCM_NEWS},
+#ifndef USE_W3MMAILER
     {"mailto", SCM_MAILTO},
+#endif
 #ifdef USE_SSL
     {"https", SCM_HTTPS},
 #endif				/* USE_SSL */
@@ -240,6 +242,7 @@ free_ssl_ctx()
     if (ssl_ctx != NULL)
 	SSL_CTX_free(ssl_ctx);
     ssl_ctx = NULL;
+    ssl_accept_this_site(NULL);
 }
 
 #if SSLEAY_VERSION_NUMBER >= 0x00905100
@@ -276,11 +279,7 @@ static SSL *
 openSSLHandle(int sock, char *hostname)
 {
     SSL *handle = NULL;
-    Str emsg;
-    Str amsg = NULL;
-    char *ans;
     static char *old_ssl_forbid_method = NULL;
-    static Str accept_this_site = NULL;
 #ifdef USE_SSL_VERIFY
     static int old_ssl_verify_server = -1;
 #endif
@@ -301,7 +300,6 @@ openSSLHandle(int sock, char *hostname)
     }
     if (ssl_path_modified) {
 	free_ssl_ctx();
-	accept_this_site = NULL;
 	ssl_path_modified = 0;
     }
 #endif				/* defined(USE_SSL_VERIFY) */
@@ -363,105 +361,15 @@ openSSLHandle(int sock, char *hostname)
 #if SSLEAY_VERSION_NUMBER >= 0x00905100
     init_PRNG();
 #endif				/* SSLEAY_VERSION_NUMBER >= 0x00905100 */
-    if (SSL_connect(handle) <= 0)
-	goto eend;
-#ifdef USE_SSL_VERIFY
-    /* check the cert chain.
-     * The chain length is automatically checked by OpenSSL when we
-     * set the verify depth in the ctx.
-     */
-    if (ssl_verify_server) {
-	X509 *x;
-	x = SSL_get_peer_certificate(handle);
-	if (x == NULL) {
-	    if (accept_this_site
-		&& strcasecmp(accept_this_site->ptr, hostname) == 0)
-		ans = "y";
-	    else {
-		emsg = Strnew_charp("No SSL peer certificate: accept (y/n)?");
-		term_raw();
-		ans = inputChar(emsg->ptr);
-	    }
-	    if (tolower(*ans) == 'y')
-		amsg =
-		    Strnew_charp
-		    ("Accept SSL session without any peer certificate");
-	    else {
-		char *e = "This SSL session was rejected "
-		    "to prevent security violation: no peer certificate";
-		disp_err_message(e, FALSE);
-		free_ssl_ctx();
-		return NULL;
-	    }
-	}
-	else {
-	    long verr;
-	    X509_free(x);
-	    if ((verr = SSL_get_verify_result(handle)) != X509_V_OK) {
-		const char *em = X509_verify_cert_error_string(verr);
-		if (accept_this_site
-		    && strcasecmp(accept_this_site->ptr, hostname) == 0)
-		    ans = "y";
-		else {
-		    emsg = Sprintf("%s: accept (y/n)?", em);
-		    term_raw();
-		    ans = inputChar(emsg->ptr);
-		}
-		if (tolower(*ans) == 'y') {
-		    amsg = Sprintf("Accept unsecure SSL session: "
-				   "unverified: %s", em);
-		}
-		else {
-		    char *e =
-			Sprintf("This SSL session was rejected: %s", em)->ptr;
-		    disp_err_message(e, FALSE);
-		    free_ssl_ctx();
-		    return NULL;
-		}
-	    }
-	}
-    }
-    else
-#endif
-	amsg = Strnew_charp("Certificate is not verified");
-
-    emsg = ssl_check_cert_ident(handle, hostname);
-    if (emsg != NULL) {
-	if (accept_this_site
-	    && strcasecmp(accept_this_site->ptr, hostname) == 0)
-	    ans = "y";
-	else {
-	    Str ep = Strdup(emsg);
-	    if (ep->length > COLS - 16)
-		Strshrink(ep, ep->length - (COLS - 16));
-	    term_raw();
-	    Strcat_charp(ep, ": accept(y/n)?");
-	    ans = inputChar(ep->ptr);
-	}
-	if (tolower(*ans) == 'y') {
-	    amsg = Strnew_charp("Accept unsecure SSL session:");
-	    Strcat(amsg, emsg);
-	}
-	else {
-	    char *e = "This SSL session was rejected "
-		"to prevent security violation";
-	    disp_err_message(e, FALSE);
-	    free_ssl_ctx();
-	    return NULL;
-	}
-    }
-    ssl_set_certificate_validity(amsg);
-    if (amsg)
-	disp_err_message(amsg->ptr, FALSE);
-    accept_this_site = Strnew_charp(hostname);
-    return handle;
+    if (SSL_connect(handle) > 0)
+	return handle;
   eend:
     close(sock);
     if (handle)
 	SSL_free(handle);
-    accept_this_site = NULL;
-    emsg = Sprintf("SSL error: %s", ERR_error_string(ERR_get_error(), NULL));
-    disp_err_message(emsg->ptr, FALSE);
+    disp_err_message(Sprintf
+		     ("SSL error: %s",
+		      ERR_error_string(ERR_get_error(), NULL))->ptr, FALSE);
     return NULL;
 }
 
@@ -626,8 +534,8 @@ openSocket(char *const hostname,
 #endif
 	goto error;
     }
-    regexCompile("^[0-9][0-9]*\\.[0-9][0-9]*\\.[0-9][0-9]*\\.[0-9][0-9]*$", 0);
-    if (regexMatch(hostname, 0, 1)) {
+    regexCompile("^[0-9]+\\.[0-9]+\\.[0-9]+\\.[0-9]+$", 0);
+    if (regexMatch(hostname, -1, 1)) {
 	sscanf(hostname, "%d.%d.%d.%d", &a1, &a2, &a3, &a4);
 	adr = htonl((a1 << 24) | (a2 << 16) | (a3 << 8) | a4);
 	bcopy((void *)&adr, (void *)&hostaddr.sin_addr, sizeof(long));
@@ -803,6 +711,10 @@ parseURL(char *url, ParsedURL *p_url, ParsedURL *current)
 	goto analyze_file;
     }
     /* scheme part has been found */
+    if (p_url->scheme == SCM_UNKNOWN) {
+	p_url->file = allocStr(url, -1);
+	return;
+    }
     /* get host and port */
     if (p[0] != '/' || p[1] != '/') {	/* scheme:foo or scheme:/foo */
 	p_url->host = NULL;
@@ -856,12 +768,6 @@ parseURL(char *url, ParsedURL *p_url, ParsedURL *current)
 	p++;
     }
     switch (*p) {
-    case '\0':
-	/* scheme://host                */
-	p_url->host = copyPath(q, -1, COPYPATH_SPC_IGNORE);
-	p_url->port = DefaultPort[p_url->scheme];
-	p_url->file = DefaultFile(p_url->scheme);
-	return;
     case ':':
 	/* scheme://user:pass@host or
 	 * scheme://host:port
@@ -881,20 +787,15 @@ parseURL(char *url, ParsedURL *p_url, ParsedURL *current)
 	/* scheme://host:port/ */
 	tmp = Strnew_charp_n(q, p - q);
 	p_url->port = atoi(tmp->ptr);
-	if (*p == '\0') {
-	    /* scheme://user@host:port      */
-	    /* scheme://user:pass@host:port */
-	    p_url->file = DefaultFile(p_url->scheme);
-	    p_url->label = NULL;
-	    return;
-	}
-	/* *p is one of ['/', '?', '#'] */
+	/* *p is one of ['\0', '/', '?', '#'] */
 	break;
     case '@':
 	/* scheme://user@...            */
 	p_url->user = copyPath(q, p - q, COPYPATH_SPC_IGNORE);
 	q = ++p;
 	goto analyze_url;
+    case '\0':
+	/* scheme://host                */
     case '/':
     case '?':
     case '#':
@@ -1046,9 +947,10 @@ parseURL2(char *url, ParsedURL *pu, ParsedURL *current)
     int relative_uri = FALSE;
 
     parseURL(url, pu, current);
+#ifndef USE_W3MMAILER
     if (pu->scheme == SCM_MAILTO)
 	return;
-
+#endif
     if (pu->scheme == SCM_LOCAL)
 	pu->file = expandName(pu->file);
 
@@ -1111,7 +1013,7 @@ parseURL2(char *url, ParsedURL *pu, ParsedURL *current)
 	    if (strncmp(pu->file, "/$LIB/", 6)) {
 		char abs[_MAX_PATH];
 
-		_abspath(abs, pu->file, _MAX_PATH);
+		_abspath(abs, file_unquote(pu->file), _MAX_PATH);
 		pu->file = file_quote(cleanupName(abs));
 	    }
 	}
@@ -1125,7 +1027,7 @@ parseURL2(char *url, ParsedURL *pu, ParsedURL *current)
 	    tmp = Strnew_charp(CurrentDir);
 	    if (Strlastchar(tmp) != '/')
 		Strcat_char(tmp, '/');
-	    Strcat_charp(tmp, pu->file);
+	    Strcat_charp(tmp, file_unquote(pu->file));
 	    pu->file = file_quote(cleanupName(tmp->ptr));
 	}
 #endif
@@ -1188,9 +1090,12 @@ _parsedURL2Str(ParsedURL *pu, int pass)
 #endif				/* USE_SSL */
     };
 
-    if (pu->scheme == SCM_UNKNOWN || pu->scheme == SCM_MISSING) {
+    if (pu->scheme == SCM_MISSING) {
 	return Strnew_charp("???");
     }
+    else if (pu->scheme == SCM_UNKNOWN) {
+	return Strnew_charp(pu->file);
+    }
     if (pu->host == NULL && pu->file == NULL && pu->label != NULL) {
 	/* local label */
 	return Sprintf("#%s", pu->label);
@@ -1205,10 +1110,12 @@ _parsedURL2Str(ParsedURL *pu, int pass)
     }
     tmp = Strnew_charp(scheme_str[pu->scheme]);
     Strcat_char(tmp, ':');
+#ifndef USE_W3MMAILER
     if (pu->scheme == SCM_MAILTO) {
 	Strcat_charp(tmp, pu->file);
 	return tmp;
     }
+#endif
 #ifdef USE_NNTP
     if (pu->scheme != SCM_NEWS)
 #endif				/* USE_NNTP */
@@ -1243,6 +1150,8 @@ _parsedURL2Str(ParsedURL *pu, int pass)
 	    )))
 	Strcat_char(tmp, '/');
     Strcat_charp(tmp, pu->file);
+    if (pu->scheme == SCM_FTPDIR && Strlastchar(tmp) != '/')
+	Strcat_char(tmp, '/');
     if (pu->query) {
 	Strcat_char(tmp, '?');
 	Strcat_charp(tmp, pu->query);
@@ -1327,28 +1236,29 @@ otherinfo(ParsedURL *target, ParsedURL *current, char *referer)
     return s->ptr;
 }
 
-static Str
-HTTPrequest(ParsedURL *pu, ParsedURL *current, HRequest *hr, TextList *extra)
+Str
+HTTPrequestMethod(HRequest *hr)
 {
-    Str tmp;
-    TextListItem *i;
-#ifdef USE_COOKIE
-    Str cookie;
-#endif				/* USE_COOKIE */
     switch (hr->command) {
     case HR_COMMAND_CONNECT:
-	tmp = Strnew_charp("CONNECT ");
-	break;
+	return Strnew_charp("CONNECT");
     case HR_COMMAND_POST:
-	tmp = Strnew_charp("POST ");
+	return Strnew_charp("POST");
 	break;
     case HR_COMMAND_HEAD:
-	tmp = Strnew_charp("HEAD ");
+	return Strnew_charp("HEAD");
 	break;
     case HR_COMMAND_GET:
     default:
-	tmp = Strnew_charp("GET ");
+	return Strnew_charp("GET");
     }
+    return NULL;
+}
+
+Str
+HTTPrequestURI(ParsedURL *pu, HRequest *hr)
+{
+    Str tmp = Strnew();
     if (hr->command == HR_COMMAND_CONNECT) {
 	Strcat_charp(tmp, pu->host);
 	Strcat(tmp, Sprintf(":%d", pu->port));
@@ -1363,6 +1273,20 @@ HTTPrequest(ParsedURL *pu, ParsedURL *current, HRequest *hr, TextList *extra)
     else {
 	Strcat(tmp, _parsedURL2Str(pu, TRUE));
     }
+    return tmp;
+}
+
+static Str
+HTTPrequest(ParsedURL *pu, ParsedURL *current, HRequest *hr, TextList *extra)
+{
+    Str tmp;
+    TextListItem *i;
+#ifdef USE_COOKIE
+    Str cookie;
+#endif				/* USE_COOKIE */
+    tmp = HTTPrequestMethod(hr);
+    Strcat_charp(tmp, " ");
+    Strcat_charp(tmp, HTTPrequestURI(pu, hr)->ptr);
     Strcat_charp(tmp, " HTTP/1.0\r\n");
     if (hr->referer == NO_REFERER)
 	Strcat_charp(tmp, otherinfo(pu, NULL, NULL));
@@ -1437,13 +1361,13 @@ openFTPStream(ParsedURL *pu)
 URLFile
 openURL(char *url, ParsedURL *pu, ParsedURL *current,
 	URLOption *option, FormList *request, TextList *extra_header,
-	URLFile *ouf, unsigned char *status)
+	URLFile *ouf, HRequest *hr, unsigned char *status)
 {
     Str tmp;
     int i, sock, scheme;
     char *p, *q, *u;
     URLFile uf;
-    HRequest hr;
+    HRequest hr0;
 #ifdef USE_SSL
     SSL *sslh = NULL;
 #endif				/* USE_SSL */
@@ -1454,6 +1378,9 @@ openURL(char *url, ParsedURL *pu, ParsedURL *current,
 #endif				/* USE_NNTP */
     int extlen = strlen(CGI_EXTENSION);
 
+    if (hr == NULL)
+	hr = &hr0;
+
     if (ouf) {
 	uf = *ouf;
     }
@@ -1491,10 +1418,10 @@ openURL(char *url, ParsedURL *pu, ParsedURL *current,
     pu->is_nocache = (option->flag & RG_NOCACHE);
     uf.ext = filename_extension(pu->file, 1);
 
-    hr.command = HR_COMMAND_GET;
-    hr.flag = 0;
-    hr.referer = option->referer;
-    hr.request = request;
+    hr->command = HR_COMMAND_GET;
+    hr->flag = 0;
+    hr->referer = option->referer;
+    hr->request = request;
 
     switch (pu->scheme) {
     case SCM_LOCAL:
@@ -1592,7 +1519,7 @@ openURL(char *url, ParsedURL *pu, ParsedURL *current,
 	    if (sock < 0)
 		return uf;
 	    uf.scheme = SCM_HTTP;
-	    tmp = HTTPrequest(pu, current, &hr, extra_header);
+	    tmp = HTTPrequest(pu, current, hr, extra_header);
 	    write(sock, tmp->ptr, tmp->length);
 	}
 	else {
@@ -1608,9 +1535,9 @@ openURL(char *url, ParsedURL *pu, ParsedURL *current,
 	if (pu->file == NULL)
 	    pu->file = allocStr("/", -1);
 	if (request && request->method == FORM_METHOD_POST && request->body)
-	    hr.command = HR_COMMAND_POST;
+	    hr->command = HR_COMMAND_POST;
 	if (request && request->method == FORM_METHOD_HEAD)
-	    hr.command = HR_COMMAND_HEAD;
+	    hr->command = HR_COMMAND_HEAD;
 	if (non_null(HTTP_proxy) &&
 	    !Do_not_use_proxy &&
 	    pu->host != NULL && !check_no_proxy(pu->host)) {
@@ -1645,20 +1572,20 @@ openURL(char *url, ParsedURL *pu, ParsedURL *current,
 #ifdef USE_SSL
 	    if (pu->scheme == SCM_HTTPS) {
 		if (*status == HTST_NORMAL) {
-		    hr.command = HR_COMMAND_CONNECT;
-		    tmp = HTTPrequest(pu, current, &hr, NULL);
+		    hr->command = HR_COMMAND_CONNECT;
+		    tmp = HTTPrequest(pu, current, hr, NULL);
 		    *status = HTST_CONNECT;
 		}
 		else {
-		    hr.flag |= HR_FLAG_LOCAL;
-		    tmp = HTTPrequest(pu, current, &hr, extra_header);
+		    hr->flag |= HR_FLAG_LOCAL;
+		    tmp = HTTPrequest(pu, current, hr, extra_header);
 		    *status = HTST_NORMAL;
 		}
 	    }
 	    else
 #endif				/* USE_SSL */
 	    {
-		tmp = HTTPrequest(pu, current, &hr, extra_header);
+		tmp = HTTPrequest(pu, current, hr, extra_header);
 		*status = HTST_NORMAL;
 		pu->label = save_label;
 	    }
@@ -1678,8 +1605,8 @@ openURL(char *url, ParsedURL *pu, ParsedURL *current,
 		}
 	    }
 #endif				/* USE_SSL */
-	    hr.flag |= HR_FLAG_LOCAL;
-	    tmp = HTTPrequest(pu, current, &hr, extra_header);
+	    hr->flag |= HR_FLAG_LOCAL;
+	    tmp = HTTPrequest(pu, current, hr, extra_header);
 	    *status = HTST_NORMAL;
 	}
 #ifdef USE_SSL
@@ -1689,7 +1616,7 @@ openURL(char *url, ParsedURL *pu, ParsedURL *current,
 		SSL_write(sslh, tmp->ptr, tmp->length);
 	    else
 		write(sock, tmp->ptr, tmp->length);
-	    if (hr.command == HR_COMMAND_POST &&
+	    if (hr->command == HR_COMMAND_POST &&
 		request->enctype == FORM_ENCTYPE_MULTIPART) {
 		if (sslh)
 		    SSL_write_from_file(sslh, request->body);
@@ -1709,7 +1636,7 @@ openURL(char *url, ParsedURL *pu, ParsedURL *current,
 		fclose(ff);
 	    }
 #endif				/* HTTP_DEBUG */
-	    if (hr.command == HR_COMMAND_POST &&
+	    if (hr->command == HR_COMMAND_POST &&
 		request->enctype == FORM_ENCTYPE_MULTIPART)
 		write_from_file(sock, request->body);
 	}
@@ -1725,7 +1652,7 @@ openURL(char *url, ParsedURL *pu, ParsedURL *current,
 	    if (sock < 0)
 		return uf;
 	    uf.scheme = SCM_HTTP;
-	    tmp = HTTPrequest(pu, current, &hr, extra_header);
+	    tmp = HTTPrequest(pu, current, hr, extra_header);
 	}
 	else {
 	    sock = openSocket(pu->host,
@@ -2048,3 +1975,144 @@ filename_extension(char *path, int is_url)
     else
 	return last_dot;
 }
+
+#ifdef USE_EXTERNAL_URI_LOADER
+static struct table2 **urimethods;
+static struct table2 default_urimethods[] = {
+    {"mailto", "file:///$LIB/w3mmail.cgi?%s"},
+    {NULL, NULL}
+};
+
+static struct table2 *
+loadURIMethods(char *filename)
+{
+    FILE *f;
+    int i, n;
+    Str tmp;
+    struct table2 *um;
+    char *up, *p;
+
+    f = fopen(expandName(filename), "r");
+    if (f == NULL)
+	return NULL;
+    i = 0;
+    while (tmp = Strfgets(f), tmp->length > 0) {
+	if (tmp->ptr[0] != '#')
+	    i++;
+    }
+    fseek(f, 0, 0);
+    n = i;
+    um = New_N(struct table2, n + 1);
+    i = 0;
+    while (tmp = Strfgets(f), tmp->length > 0) {
+	if (tmp->ptr[0] == '#')
+	    continue;
+	while (IS_SPACE(Strlastchar(tmp)))
+	    Strshrink(tmp, 1);
+	for (up = p = tmp->ptr; *p != '\0'; p++) {
+	    if (*p == ':') {
+		um[i].item1 = Strnew_charp_n(up, p - up)->ptr;
+		p++;
+		break;
+	    }
+	}
+	if (*p == '\0')
+	    continue;
+	while (*p != '\0' && IS_SPACE(*p))
+	    p++;
+	um[i].item2 = Strnew_charp(p)->ptr;
+	i++;
+    }
+    um[i].item1 = NULL;
+    um[i].item2 = NULL;
+    fclose(f);
+    return um;
+}
+
+void
+initURIMethods()
+{
+    TextList *methodmap_list = NULL;
+    TextListItem *tl;
+    int i;
+
+    if (non_null(urimethodmap_files))
+	methodmap_list = make_domain_list(urimethodmap_files);
+    if (methodmap_list == NULL)
+	return;
+    urimethods = New_N(struct table2 *, (methodmap_list->nitem + 1));
+    for (i = 0, tl = methodmap_list->first; tl; tl = tl->next) {
+	urimethods[i] = loadURIMethods(tl->ptr);
+	if (urimethods[i])
+	    i++;
+    }
+    urimethods[i] = NULL;
+}
+
+Str
+searchURIMethods(ParsedURL *pu)
+{
+    struct table2 *ump;
+    int i;
+    Str scheme = NULL;
+    Str url;
+    char *p;
+
+    if (pu->scheme != SCM_UNKNOWN)
+	return NULL;		/* use internal */
+    if (urimethods == NULL)
+	return NULL;
+    url = parsedURL2Str(pu);
+    for (p = url->ptr; *p != '\0'; p++) {
+	if (*p == ':') {
+	    scheme = Strnew_charp_n(url->ptr, p - url->ptr);
+	    break;
+	}
+    }
+    if (scheme == NULL)
+	return NULL;
+
+    for (i = 0; (ump = urimethods[i]) != NULL; i++) {
+	for (; ump->item1 != NULL; ump++) {
+	    if (strcmp(ump->item1, scheme->ptr) == 0) {
+		return Sprintf(ump->item2, url_quote(url->ptr));
+	    }
+	}
+    }
+    for (ump = default_urimethods; ump->item1 != NULL; ump++) {
+	if (strcmp(ump->item1, scheme->ptr) == 0) {
+	    return Sprintf(ump->item2, url_quote(url->ptr));
+	}
+    }
+    return NULL;
+}
+
+/*
+ * RFC2396: Uniform Resource Identifiers (URI): Generic Syntax
+ * Appendix A. Collected BNF for URI
+ * uric          = reserved | unreserved | escaped
+ * reserved      = ";" | "/" | "?" | ":" | "@" | "&" | "=" | "+" |
+ *                 "$" | ","
+ * unreserved    = alphanum | mark
+ * mark          = "-" | "_" | "." | "!" | "~" | "*" | "'" |
+ *                  "(" | ")"
+ * escaped       = "%" hex hex
+ */
+
+#define URI_PATTERN	"([-;/?:@&=+$,a-zA-Z0-9_.!~*'()]|%[0-9A-Fa-f][0-9A-Fa-f])*"
+void
+chkExternalURIBuffer(Buffer *buf)
+{
+    int i;
+    struct table2 *ump;
+
+    for (i = 0; (ump = urimethods[i]) != NULL; i++) {
+	for (; ump->item1 != NULL; ump++) {
+	    reAnchor(buf, Sprintf("%s:%s", ump->item1, URI_PATTERN)->ptr);
+	}
+    }
+    for (ump = default_urimethods; ump->item1 != NULL; ump++) {
+	reAnchor(buf, Sprintf("%s:%s", ump->item1, URI_PATTERN)->ptr);
+    }
+}
+#endif
